# Pyslk Readme

This is a python wrapper for `slk` and `slk_helpers`. `slk` is the command line interface of StrongLink, which is the software of the new HSM system at DKRZ. `slk_helpers` are an in-house development of DKRZ which extends `slk`. This wrapper is compatible with slk versions `>= 3.3.76` (`>= 3.3.90` recommended) and slk helpers versions `>= 1.12.0` (`>= 1.12.10` recommended).

If you have any questions, please contact support@dkrz.de .


## Installation

The installation requires the following dependencies

* `>= python 3.9`
* `>= slk 3.3.76` (`>= slk 3.3.90` recommended)
* `>= slk_helpers 1.12.0` (`>= slk_helpers 1.12.10` recommended)
* `pandas`
* `psutil`
* `tenacity`


The package can be installed via pip:

```console
pip install pyslk -U --extra-index-url https://gitlab.dkrz.de/api/v4/projects/1848/packages/pypi/simple
```

### Installation with conda/mamba

You can also install `pyslk` into a conda environment:

```console
mamba create -n slk pip
conda activate slk
```

We recommend to install the package via `pip` in your conda environment because there exists `pip` / `PyPi` download URL which automatically installs the newest release from the hosting GitLab repository. If you want to install the package via conda, please do as follows:

* download the latest release (2.0.0) for your Python version from one of these locations:
** https://gitlab.dkrz.de/hsm-tools/pyslk/-/packages
** https://hsm-tools.gitlab-pages.dkrz.de/pyslk/availability.html#conda-packages
* install the package es follows (example for pyslk 2.0.0 with Python 3.11): `conda install --offline  pyslk-2.0.0-py311_0.tar.bz2`

Run install with pip (latest release):

```console
pip install pyslk -U --extra-index-url https://gitlab.dkrz.de/api/v4/projects/1848/packages/pypi/simple
```


## Usage

All command lines arguments of `slk` v3.3.91 and `slk_helpers` v1.12.10 are available as the functions' arguments.

A few basic usage examples:

```python
# load library
> import pyslk
> import os
# list content of a folder
> pyslk.list("/arch/bm0146/k204221/iow")
    permissions    owner   group      filesize       date   time                                    filename
0   -rwxr-xr-x-  k204221  bm0146  1.268945e+06 2020-06-10  08:25          /arch/bm0146/k204221/iow/INDEX.txt
1   -rw-r--r--t  k204221  bm0146  2.094216e+10 2020-06-05  17:36  /arch/bm0146/k204221/iow/iow_data2_001.tar
2   -rw-r--r--t  k204221  bm0146  2.034971e+10 2020-06-05  17:38  /arch/bm0146/k204221/iow/iow_data2_002.tar
3   -rw-r--r--t  k204221  bm0146  2.088344e+10 2020-06-05  17:38  /arch/bm0146/k204221/iow/iow_data2_003.tar
4   -rw-r--r--t  k204221  bm0146  2.071567e+10 2020-06-05  17:40  /arch/bm0146/k204221/iow/iow_data2_004.tar
5   -rw-r--r--t  k204221  bm0146  2.047869e+10 2020-06-05  17:40  /arch/bm0146/k204221/iow/iow_data2_005.tar
6   -rw-r--r--t  k204221  bm0146  8.364491e+09 2020-06-05  17:41  /arch/bm0146/k204221/iow/iow_data2_006.tar
7   -rw-r--r--t  k204221  bm0146  2.007016e+11 2020-06-05  19:37  /arch/bm0146/k204221/iow/iow_data3_001.tar
8   -rw-r--r--t  k204221  bm0146  2.646606e+10 2020-06-05  19:14  /arch/bm0146/k204221/iow/iow_data3_002.tar
9   -rw-r--r--t  k204221  bm0146  4.194304e+06 2020-06-05  19:43  /arch/bm0146/k204221/iow/iow_data4_001.tar
10  -rw-r--r--t  k204221  bm0146  1.128477e+10 2020-06-05  19:46  /arch/bm0146/k204221/iow/iow_data4_002.tar
11  -rw-r--r--t  k204221  bm0146  2.094216e+10 2020-06-10  08:21  /arch/bm0146/k204221/iow/iow_data5_001.tar
12  -rw-r--r--t  k204221  bm0146  2.034971e+10 2020-06-10  08:23  /arch/bm0146/k204221/iow/iow_data5_002.tar
13  -rw-r--r--t  k204221  bm0146  2.088344e+10 2020-06-10  08:23  /arch/bm0146/k204221/iow/iow_data5_003.tar
14  -rw-r--r--t  k204221  bm0146  2.071567e+10 2020-06-10  08:24  /arch/bm0146/k204221/iow/iow_data5_004.tar
15  -rw-r--r--t  k204221  bm0146  2.047869e+10 2020-06-10  08:25  /arch/bm0146/k204221/iow/iow_data5_005.tar
16  -rw-r--r--t  k204221  bm0146  8.364491e+09 2020-06-10  08:25  /arch/bm0146/k204221/iow/iow_data5_006.tar
17  -rw-r--r--t  k204221  bm0146  2.094216e+10 2020-06-05  17:53   /arch/bm0146/k204221/iow/iow_data_001.tar
18  -rw-r--r--t  k204221  bm0146  2.034971e+10 2020-06-05  17:53   /arch/bm0146/k204221/iow/iow_data_002.tar
19  -rw-r--r--t  k204221  bm0146  2.088344e+10 2020-06-05  17:56   /arch/bm0146/k204221/iow/iow_data_003.tar
20  -rw-r--r--t  k204221  bm0146  2.071567e+10 2020-06-05  17:56   /arch/bm0146/k204221/iow/iow_data_004.tar
21  -rw-r--r--t  k204221  bm0146  2.047869e+10 2020-06-05  17:58   /arch/bm0146/k204221/iow/iow_data_005.tar
22  -rw-r-----t  k204221  bm0146  8.364491e+09 2020-06-05  17:57   /arch/bm0146/k204221/iow/iow_data_006.tar

# check if a file exists
> pyslk.resource_exists("/arch/bm0146/k204221/iow/iow_data2_001.tar")
True
# check another non-existing file
> pyslk.resource_exists("/arch/bm0146/k204221/iow/NOT_EXISTS.tar")
False
# get metadata of a file
> pyslk.get_metadata("/arch/bm0146/k204221/iow/INDEX.txt")
{'netcdf.Title': 'ABC DEF GHI'}
# get size
> pyslk.get_resource_size("/arch/bm0146/k204221/iow/INDEX.txt")
1268945
# get checksums
> pyslk.get_checksum("/arch/bm0146/k204221/iow/INDEX.txt")
{'adler32': '412dec3a',
 'sha512': '83644ffc384c70aaf60d29f3303d771e219f3b3f8abdb3a590ac501faf7abbeb371ee4b6b949a1cdc3fa7c44438cd6643f877e8c1ab27d8fe5cc26eb06896d99'}
# is the file stored in the cache? yet!
> pyslk.is_cached("/arch/bm0146/k204221/iow/INDEX.txt")
True
# retrieve file
> pyslk.retrieve("/arch/bm0146/k204221/iow/INDEX.txt", '.', preserve_path=False)
''
> os.path.exists('INDEX.txt')
True
```

We want to search for files, which are stored in ``/arch/bm0146/k204221/iow``, are larger than 1 kb and smaller
than approximately 2 MB.

```python
# load pyslk
import pyslk.core.gen_queries
import pyslk.base.searching
> import pyslk
# generate search query
> search_query = pyslk.gen_search_query(["path=/arch/bm0146/k204221/iow", "resources.size>1024", "resources.size<2000000"])
> search_query
'{"$and":[{"resources.size":{"$gt":1024}},{"resources.size":{"$lt":2000000}},{"path":{"$gte":"/arch/bm0146/k204221/iow","$max_iterations":1}}]}'
# run search
> search_id = pyslk.search(search_query)
> search_id
677761
# list search results
> pyslk.list(search_id)
   permissions    owner   group   filesize       date   time                            filename
0  -rwxr-xr-x-  k204221  bm0146  1268945.0 2020-06-10  08:25  /arch/bm0146/k204221/iow/INDEX.txt
```

or just

```python
> import pyslk
> pyslk.list(pyslk.search(pyslk.core.gen_queries.gen_search_query(["path=/arch/bm0146/k204221/iow", "resources.size>1024", "resources.size<2000000"])))
   permissions    owner   group   filesize       date   time                            filename
0  -rwxr-xr-x-  k204221  bm0146  1268945.0 2020-06-10  08:25  /arch/bm0146/k204221/iow/INDEX.txt
```

Split one large retrieval request into multiple retrieval requests and run them

```python
> import pprint
> import pyslk
# create grouping
> grouping = pyslk.group_files_by_tape('/arch/bm0146/k204221/iow', recursive=True, run_search_query=True)
# look into the grouping
> pprint.pprint(grouping)
[{'barcode': '',
  'description': 'files currently stored in the HSM cache',
  'file_count': 3,
  'file_ids': [49058705518, 49058705517, 49058705497],
  'files': ['/arch/bm0146/k204221/iow/iow_data_005.tar',
            '/arch/bm0146/k204221/iow/iow_data_004.tar',
            '/arch/bm0146/k204221/iow/INDEX.txt'],
  'id': -1,
  'location': 'cache',
  'search_complete': True,
  'search_id': 505589,
  'search_query': '{"$and":[{"path":{"$gte":"/arch/bm0146/k204221/iow"}},{"smart_pool":"slpstor"}]}',
  'status': ''},
 ...
 {'barcode': 'M12217M8',
  'description': 'files stored on the tape with the given ID and barcode/label',
  'file_count': 2,
  'file_ids': [49058705503, 49058705502],
  'files': ['/arch/bm0146/k204221/iow/iow_data2_006.tar',
            '/arch/bm0146/k204221/iow/iow_data2_005.tar'],
  'id': 130879,
  'location': 'tape',
  'search_complete': True,
  'search_id': 505599,
  'search_query': '{"$and":[{"path":{"$gte":"/arch/bm0146/k204221/iow"}},{"tape_barcode":"M12217M8"}]}',
  'status': 'AVAILABLE'}]
# print the search ids 
> [r['search_id'] for r in grouping]
[505589, 505590, 505591, 505592, 505593, 505594, 505595, 505596, 505597, 505598, 505599]
# run retrievals
> pyslk.retrieve(505589, '.', preserve_path=True)
> pyslk.retrieve(505590, '.', preserve_path=True)
> pyslk.retrieve(505591, '.', preserve_path=True)
> pyslk.retrieve(505592, '.', preserve_path=True)
> pyslk.retrieve(505593, '.', preserve_path=True)
> pyslk.retrieve(505594, '.', preserve_path=True)
> pyslk.retrieve(505595, '.', preserve_path=True)
> pyslk.retrieve(505596, '.', preserve_path=True)
> pyslk.retrieve(505597, '.', preserve_path=True)
> pyslk.retrieve(505598, '.', preserve_path=True)
> pyslk.retrieve(505599, '.', preserve_path=True)
```


### `pyslk` in the DKRZ jupyterhub

To make use of `pyslk` in the jupyterhub, you have to make the `slk` executable
available in the `PATH` environment variable. You can do this by creating your own
`kernel.json` file, e.g.,

```console
module load slk
python -m ipykernel install --name slk --display-name "hsm kernel" --env PATH $PATH --env JAVA_HOME $JAVA_HOME --user
```

You need to restart the kernel if you modified/replaced an existing `kernel.json`. You can find
your `kernel.json`, e.g., in `.local/share/jupyter/kernels/your_kernel` depending
on where you store your kernel files (probably in your `$HOME` directory).

If you want to use a different version of `slk`, you have to update the path to your executable.

See also [here](https://jupyterhub.gitlab-pages.dkrz.de/jupyterhub-docs/posts/2020/nosuchfile.html).

## Contributing
Anyone is welcome to contribute to the code. To do so you must clone the
content of this repository and install the library using the following command:

```console
pip install -e .[tests]
```

This will install all requirements that are necessary for local development.
It is best practice to create a new branch before adding any changes.

```console
git checkout -b my-meaninful-branchname
```

After all changes have been committed and pushed back to the repository you can
submit the merge request that will be reviewed by the slk team.

### Code conventions
The gitlab test pipeline performs format checking. To check if the code is
still formatted correctly you can add a pre-commit hook that performs those
checks every time you are about to commit changes to the code. To add the
commit hook run the following command:

```console
pre-commit install
```

### Prepare and create a release (developers only)

run tests:
```console
pytest -v tests
```

check codestyle:
```console
flake8 pyslk
black --check -t py310 pyslk
isort --check --profile black pyslk
```

update version number where needed and write changelog (for example release 2.0.1)

```console
vim CHANGES.md         # update changes for the new version (for GitLab Repo Readme)
vim doc/changelog.rst  # update changes for the new version (for online documentation)
# increment version number from 2.0.0 to 2.0.1
sed -i 's/2\.0\.1/2.0.2/g' doc/development.rst README.md
sed -i 's/2\.0\.0/2.0.1/g' pyproject.toml meta.yaml doc/availability.rst doc/development.rst README.md
# manually imcrement the 'sed' expressions from the lines above in README.MD and doc/development.rst
vim README.MD
vim doc/development.rst
# set tag and commit in git
git tag 2.0.1        # tag version
git commit -a        # commit changes
git push --tags      # push changes including tag
git push             # push changes
```


make a release for PyPI (example for version 2.0.1):

```console
python -m pip install build
python -m build . --sdist
# twine upload dist/pyslk-2.0.1.tar.gz     # upload to pypi
```

make a conda package (example for version 2.0.1):

```console
conda-build .          # build conda packages
mkdir conda            # create folder to move conda packages to
# move conda packages
cp /opt/python/anaconda3/conda-bld/linux-64/pyslk-2.0.1-py39_0.tar.bz2 conda/
cp /opt/python/anaconda3/conda-bld/linux-64/pyslk-2.0.1-py310_0.tar.bz2 conda/
cp /opt/python/anaconda3/conda-bld/linux-64/pyslk-2.0.1-py311_0.tar.bz2 conda/
cp /opt/python/anaconda3/conda-bld/linux-64/pyslk-2.0.1-py312_0.tar.bz2 conda/
```

### Generate documentation (developers only)

```console
# install needed dependencies for doc generation 
pip3 install -U mock sphinx-book-theme numpydoc
# install this package without dependencies
pip3 install --no-deps -e .
# If dependencies are missing, please check whether all dependencies, which are listed in section
# `project.optional-dependencies` under `doc`, are also installed here. I dependencies are missing here, please add
# them. 

# Generate `pyslk.rst` and `modules.rst` files which contain only `:orphan:`. These files will be ignored by Sphinx
# during the generation of the documentation. If they would not exist, sphinx-autodoc would generate them/  
if [ ! -e doc/pyslk.rst ]; then echo ":orphan:" > doc/pyslk.rst; fi
if [ ! -e doc/modules.rst ]; then echo ":orphan:" > doc/modules.rst; fi
# Prepare doc generation
sphinx-apidoc -o doc pyslk
# Generate documentation
sphinx-build -b html doc build/html
# you will find the documentation in `./build/html/index.txt`
```

> **_NOTE_**: Before generating the documentation the first time, the `doc/conf.py` was edited as follows: `sys.path.insert(0, os.path.abspath('..'))` was set and `'sphinx.ext.autodoc'` was inserted in the list `extensions`.


## Acknowledgements

Lars Buntemeyer (GERICS), Carsten Ehbrecht (DKRZ), Andrej Fast (DKRZ), Helge Heuer (DLR), Karsten Peters-von Gehlen (DKRZ), Fabian Wachsmann (DKRZ) and Karl-Hermann Wieners (MPI-M) contributed to this package. They are listed in alphabetical order.
